@@ -1,7 +1,7 @@
 # -*- encoding: utf-8 -*-
 # vim: expandtab:tabstop=4:softtabstop=4:autoindent
 
-"""nmevent - C#-like implementation of the Observer pattern
+"""nmevent v0.2 - C#-like implementation of the Observer pattern
 
 This is a Python module :mod:`nmevent`, simple C#-like implementation of
 the Observer pattern (http://en.wikipedia.org/wiki/Observer_pattern).
@@ -24,22 +24,17 @@ The most straightfoward way to use :mod:`nmevent` is this:
 >>> example = ExampleClass()
 >>> example.event += handler
 
-It should be noted, that event doesn't necessarily need to
-be an object attribute. :class:`Event` instance is basically
-just a callable object that works as a sort of "dispatch
-demultiplexer".
+It should be noted, that event doesn't necessarily need to be an object
+attribute. :class:`Event` instance is basically just a callable object that
+works as a sort of "dispatch demultiplexer".
 
-This usage, however, isn't very C#-like. In C#, events are declared
-in class scope and that's where the :class:`EventSlot` comes in.
-Once you've created event using :class:`EventSlot`, you can
-use the same way you use :class:`Event`, only you don't need
-to specify the sender when raising the event. That's because
-the event is already bound to the instance of the class it has
-been declared in.
+This usage, however, isn't very C#-like. In C#, events are declared in class
+scope and that's why the :class:`Event` class also supports the descriptor
+protocol (you can use the same way you use the built-in ``property`` object).
 
 >>> from nmevent import EventSlot
 >>> class ExampleClass(object):
-...    event = EventSlot()
+...    event = Event()
 ...
 ...    def _do_something(self):
 ...       self.event()
@@ -50,9 +45,19 @@ been declared in.
 >>> example = ExampleClass()
 >>> example.event += handler
 
-Perhaps this looks even more straightfoward than instantiating
-:class:`Event` in object's constructor, but there's actually
-lot more going on under hood this time.
+Perhaps this looks even more straightfoward than instantiating :class:`Event`
+in object's constructor, but there's actually lot more going on under hood this
+time.
+
+Finally, there is the :class:`Property` descriptor and the associated
+:func:`nmproperty` function decorator, which work very much like the built-in
+``property`` object and decorator, except it can optionally call a callback
+function if the property's value changes after calling the setter function. It
+can work in tandem with the :func:`with_events` class decorator, which
+decorates the class with property change events and connects them to the
+instances of :class:`Property` class. It also creates events for the built-in
+``property`` objects, but you have to raise the events yourself in the setter
+function or elsewhere.
 
 >>> import nmevent
 >>> @nmevent.with_events
@@ -61,6 +66,19 @@ lot more going on under hood this time.
 ...    def x(self):
 ...       return self._x
 ...
+...    @x.setter
+...    def x(self, value):
+...       self._x = value
+...
+...    @property
+...    def y(self):
+...       return self._y
+...
+...    @y.setter
+...    def y(self, value):
+...       old_value, self._y = self._y, value
+...       self.y_changed(old_value = old_value)
+...
 >>> example = ExampleClass()
 >>> example.x_changed += handler
 >>> example.x = 10 # handler gets called
@@ -84,27 +102,69 @@ You should have received a copy of the GNU Lesser General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 """
 
+__version__ = __doc__.splitlines()[0].split(' ')[1][1:]
 __author__ = u"Jan Milik <milikjan@fit.cvut.cz>"
 __all__    = [
     'nmproperty',
     'with_events',
-    'EventSlot',
     'Event',
-    'EventArgs',
 ]
 
 import __builtin__
 
-class EventArgs(object):
-    """Base class for event arguments objects.
-    """
+EVENTS_ATTRIBUTE = '__nmevents__'
 
-    def __init__(self, **keywords):
-        self.keywords = keywords
+class CallbackStore(object):
+    """Collection of callbacks."""
 
-    def __getattr__(self, name):
-        if name in self.keywords:
-            return self.keywords[name]
+    def __init__(self):
+        """Constructor."""
+        self.callbacks = set()
+
+    def __iter__(self):
+        """Returns the collection's iterator object."""
+        return iter(self.callbacks)
+
+    def add(self, callback):
+        """Adds a callback to callection.
+
+        :param callback: callable object to be added
+        """
+        self.callbacks.add(callback)
+        return self
+    __iadd__ = add
+
+    def remove(self, callback):
+        """Removes a callback from the collection.
+
+        :param callback: callback to be removed
+        """
+        self.callbacks.remove(callback)
+        return self
+    __isub__ = remove
+
+    def contains(self, callback):
+        """Returns ``True`` is ``callback`` is in the collection.
+
+        :param callback: callback to check for
+        """
+        return callback in self.callbacks
+    __contains__ = contains
+
+    def count(self):
+        """Returns the number of callbacks in the collection."""
+        return len(self.callbacks)
+    __len__ = count
+
+    def clear(self):
+        """Removes all callbacks from collection."""
+        self.callbacks = set()
+
+    def call(self, *args, **keywords):
+        """Calls all callbacks with the given arguments."""
+        for callback in self.callbacks:
+            callback(*args, **keywords)
+    __call__ = call
 
 class Event(object):
     """Subject in the observer pattern.
@@ -124,11 +184,41 @@ class Event(object):
     ...
     """
 
+    __slots__ = ('__handlers__', )
+
+    @property
+    def handlers(self):
+        """Collection of this event's handlers."""
+        if self.__handlers__ is None:
+            self.__handlers__ = CallbackStore()
+        return self.__handlers__
+        
     def __init__(self):
-        self.handlers = set()
+        self.__handlers__ = None
+
+    def __get__(self, obj, objtype = None):
+        return self.bind(objtype, obj)
+
+    def __set__(self, obj, value):
+        # raise AttributeError, "Events are read-only attributes."
+        pass
+
+    def __delete__(self, obj):
+        raise AttributeError, "Events are read-only attributes."
+
+    def bind(self, objtype, obj = None):
+        """Binds the event to a class and optionally an instance."""
+        return InstanceEvent(self, objtype, obj)
     
     def add_handler(self, handler):
-        """Adds a handler (observer) to this event."""
+        """Adds a handler (observer) to this event.
+
+        ``__iadd__`` attribute of this class is just an alias of this
+        method, so the two following statements are equivalent:
+
+        >>> event.add_handler(handler)
+        >>> event += handler
+        """
         self.handlers.add(handler)
         return self
     __iadd__ = add_handler
@@ -155,108 +245,119 @@ class Event(object):
     def fire(self, sender, *args, **keywords):
         """Fires this event and calls all of its handlers.
         """
-        for handler in self.handlers:
-            handler(sender, *args, **keywords)
+        self.handlers.call(sender, *args, **keywords)
     __call__ = fire
     
     def disconnect(self):
         """Disconnects this event from all handlers.
         """
-        self.handlers = set()
+        self.handlers.clear()
 
-class BoundEvent(object):
-    """Event bound to an instance of an object.
+class InstanceEvent(object):
+    """Bound or unbound event.
 
-    Bound event automatically uses the object it is
-    bound to as its sender when raising the event.
-    This class is not to be used directly by the client
-    code; instances of this class are created by
-    the :class:`EventSlot` class.
+    In Python, unbound actually means bound to a class.
+    Bound means bound to both a class and an instance.
+    Instances of this class cannot be bound to the 
+    ``None`` object, because it is used to indicate
+    that the event is unbound.
 
-    The difference between :class:`Event` and :class:`EventSlot`
-    is somewhat similar to the difference between python's
-    functions and bound methods (bound method automatically
-    passes the object it is bound to to the function it envelopes
-    as the first argument).
+    This class is meant to be instantiated either through
+    the :class:`Event`'s descriptor protocol, or by
+    the :meth:`Event.bind` method.
 
-    Usage:
+    :param event: :class:`Event` instance to be bound
+    :param clss: class the event should be bound to (sender must be of this type)
+    :param sender: sender to bind the event to
 
-    >>> instance += handler
-    >>> instance -= handler
-    >>> handler in instance
-    >>> instance(*args, **keywords)
+    .. attribute:: __slots__
+       
+       This class uses the ``__slots__`` attribute to save
+       memory, so don't try to assign new attributes to
+       its instances.
+
+    .. attribute:: im_event
+       
+       :class:`Event` instance that is bound.
+
+    .. attribute:: im_class
+
+       Class object this event is bound to.
+
+    .. attribute:: im_sender
+
+       Instance this event is bound to.
+
+       The following condition must be always true:
+
+       >>> isinstance(self.im_sender, self.im_class)
     """
+    
+    __slots__ = ('im_event', 'im_class', 'im_sender', )
 
-    def __init__(self, obj, event):
-        self.obj   = obj
-        self.event = event
+    @property
+    def is_bound(self):
+        """``True`` if the event is bound to a sender, ``False`` otherwise."""
+        return bool(self.im_sender is not None)
 
-    def __getattr__(self, name):
-        return getattr(self.event, name)
+    @property
+    def handlers(self):
+        """:class:`CallbackStore` object that stores this event's handlers."""
+        if not self.is_bound:
+            return self.im_event.handlers
+        
+        sender = self.im_sender
+        events = sender.__dict__.setdefault(EVENTS_ATTRIBUTE, {})
+        handlers = events.setdefault(id(self.im_event), CallbackStore())
+        return handlers
+
+    def __init__(self, event, clss, sender = None):
+        self.im_event = event
+        self.im_class = clss
+        self.im_sender = sender
+    
+    def __call__(self, *args, **keywords):
+        sender = self.im_sender
+        if sender is None:
+            if len(args) < 1:
+                raise TypeError, ("Unbound event must be called with "
+                    "at least 1 positional argument representing the sender.")
+            if type(args[0]) is not self.im_class:
+                raise TypeError, ("This unbound event must be called with "
+                    "%s instance as the first argument." % 
+                        (self.im_class.__name__))
+            sender = args[0]
+            args = args[1:]
+        return self.handlers(sender, *args, **keywords)
 
     def __iadd__(self, handler):
-        self.event += handler
+        if self.is_bound:
+            self.handlers.add(handler)
+        else:
+            self.im_event += handler
         return self
 
     def __isub__(self, handler):
-        self.event -= handler
+        if self.is_bound:
+            self.handlers.remove(handler)
+        else:
+            self.im_event -= handler
         return self
 
     def __contains__(self, handler):
-        return (handler in self.event)
+        if self.is_bound:
+            return handler in self.handlers
+        else:
+            return handler in self.im_event
 
-    def __call__(self, *args, **keywords):
-        self.event(self.obj, *args, **keywords)
-
-class EventSlot(object):
-    """Event descriptor.
-
-    This is an event descriptor, which means it works
-    kind of like the built-in property object. You can
-    find more information on descriptors in Python's
-    documentation (http://docs.python.org/reference/datamodel#customizing-attribute-access),
-    or in a number of articles such as Raymond Hettinger's
-    how-to (http://users.rcn.com/python/download/Descriptor.htm).
-    In short, as the Python documentation puts it, descriptor
-    is an object with "binding behavior".
-
-    The rationale for this class is that, unlike the :class:`Event`
-    class alone, it alows client to declare events in the class scope.
-    The events don't need the ``__init__()`` method to be run in order
-    to be instantiated. This is especially helpfull when using inheritance,
-    because the inherited class doesn't need to explicitly call base
-    class's constructor in order to inherit the events.
-
-    Also, declaring the events in the class scope is more C#-like
-    than instantiating them in the constructor.
-
-    Usage:
-
-    >>> class Example(object):
-    ...    something_changed = EventSlot()
-    ...
-    ...    def foo(self):
-    ...       self.something_changed()
-    ...
-    >>> example = Example()
-    >>> example.something_changed += handler
-    """
-
-    EVENTS_ATTRIBUTE = '__events__'
-
-    def __get__(self, obj, objtype = None):
-        if obj is None:
-            return self
-        events = obj.__dict__.get(self.EVENTS_ATTRIBUTE, None)
-        if events is None:
-            events = {}
-            obj.__dict__[self.EVENTS_ATTRIBUTE] = events
-        event = events.get(id(self), None)
-        if event is None:
-            event = Event()
-            events[id(self)] = event
-        return BoundEvent(obj, event)
+    def __getattr__(self, name):
+        return getattr(self.im_event, name)
 
+    def __str__(self):
+        if self.is_bound:
+            return "<bound event>"
+        return "<unbound event>"
+    
 class Property(object):
     """Eventful property descriptor.
 
@@ -375,22 +476,16 @@ class Property(object):
         old_value = self.fget(obj)
         self.fset(obj, value)
         if old_value != value:
-            self._fire(self.changed, obj, old_value = old_value)
-            self._fire(self.property_changed, obj, old_value = old_value, name = self.name)
+            if self.changed is not None:
+                self.changed(old_value = old_value)
+            if self.property_changed is not None:
+                self.property_changed(old_value = old_value, name = self.name)
 
     def __delete__(self, obj):
         if self.fdel is None:
             raise AttributeError, "Can't delete attribute."
         self.fdel(obj)
 
-    def _fire(self, event, obj, **keywords):
-        """Propety._fire(event, obj, **keywords) - private helper function.
-        """
-        if isinstance(event, EventSlot):
-            event.__get__(obj)(**keywords)
-        elif isinstance(event, Event):
-            event(obj, **keywords)
-
     def setter(self, function):
         """Sets the setter function and returns self.
 
@@ -424,7 +519,7 @@ class Property(object):
         return self
 
     def deleter(self, function):
-        """Property.deleter(function) - method decorator te set the delete function.
+        """Method decorator to set the delete function.
 
         Works exatcly like the built-in @property.deleter.
         
@@ -440,8 +535,7 @@ def nmproperty(function):
     Creates new :class:`Property` object using the decorated method
     as the getter function. Setter and deleter functions can be
     set by the :meth:`Property.setter` and :meth:`Property.deleter`
-    decorators where "name" is the name of the property (the name
-    of the getter function).
+    decorators.
 
     This decorator is called :func:`nmproperty` to avoid name conflict
     with the built-in `property` function and decorator.
@@ -464,8 +558,8 @@ def nmproperty(function):
     >>> example.x_changed += handler # "handler" will be called when the value of x changes
     >>> example.x = 10 # value of x changed, "handler" should get called
 
-    The changed events can be automatically created and set
-    by the :func:`with_events` decorator.
+    The :attr:`Property.changed` events can be automatically created and set
+    by the :func:`with_events` decorator when used on the class.
 
     :param function: function to be used as the property getter function
     :returns: new `Property` object
@@ -475,22 +569,60 @@ def nmproperty(function):
 def with_events(clss):
     """Decorates a class with some automatic event slots.
 
-    Automatically adds property change notification events
-    of the name "x_changed", where x is the name of the
-    property.
-
     :param clss: class object to be decorated
-    :returns: decorated class
+    :returns:    decorated class
+
+    Automatically adds property change notification events of the name
+    "x_changed", where x is the name of the property.
+
+    Usage:
+
+    >>> @nmevent.with_events
+    ... class Example(object):
+    ...    @nmevent.nmproperty
+    ...    def x(self):
+    ...       return self._x
+    ...
+    ...    @x.setter
+    ...    def x(self, value):
+    ...       self._x
+    ...
+    ...    def __init__(self):
+    ...       self._x = 0
+    ...
+    >>> def x_changed_handler(self, sender, **keywords):
+    ...    old_value = keywords['old_value']
+    ...    print "x changed; %r -> %r" % (old_value, sender.x)
+    ...
+    >>> def property_changed_handler(self, sender, **keywords):
+    ...    old_value = keywords['old_value']
+    ...    name = keywords['name']
+    ...    print "property \"%s\" changed, %r -> %r" % (old_value, sender.x)
+    ...
+    >>> example = Example()
+    >>> example.x_changed += x_changed_handler
+    >>> example.property_changed_handler += property_changed_handler
+    >>> example.x = 42
+
+    In the example above, the :func:`with_events` decorator automatically
+    decorates the class with an ``x_changed`` event and ``property_changed``
+    event connects them to the instance of :class:`Property` class created by
+    the :func:`nmproperty` decorator.
+
+    Simply put, the class has ``x_changed`` event and ``property_changed``
+    events that are raised when the value of ``Example.x`` changes.
+    ``x_changed`` gets called only when ``Example.x`` changes,
+    ``property_changed`` gets called when any property changes.
     """
 
-    property_changed = EventSlot()
+    property_changed = Event()
     setattr(clss, "property_changed", property_changed)
 
     for name, attr in clss.__dict__.items():
         if isinstance(attr, __builtin__.property):
-            setattr(clss, name + "_changed", EventSlot())
+            setattr(clss, name + "_changed", Event())
         elif isinstance(attr, Property):
-            slot = EventSlot()
+            slot = Event()
             setattr(clss, name + "_changed", slot)
             attr.changed = slot
             attr.property_changed = property_changed
